;; -*- lexical-binding: t -*-


(defun my-print-table (table &optional col-formatters alignments col-separator
                             header-line)
  "Return formatted TABLE, a matrix of strings, as a string.

TABLE is a list of lists of strings which is interpreted as a
list of table lines.

COL-FORMATTERS is a list of functions (or nil entries) to format
the according column entries.  These functions should all accept
one string as argument, an table entry to print.

ALIGNMENTS is a list of elements of the form
[left|right|centered|nil].

COL-FORMATTERS and ALIGNMENTS are allowed to be shorter than the
number of entries in TABLE.

COL-SEPARATOR is a string to use as a column separator, like
\"|\".  Default is \" \".  It can alternatively be a list of
number of columns minus 1 separators to use.

HEADER-LINE is an additional table line to use as header line.
Should have as many elements as the table has columns.

Example:

  (insert
   (my-print-table
    (let ((numbers (number-sequence 1 15)))
      (mapcar (lambda (n) (list n (expt n 2) (sqrt n))) numbers))
    (list #\\='number-to-string
          #\\='number-to-string
          (apply-partially #\\='format \"%.5f\"))
    (list \\='right \\='right \\='center)
    \" | \"
    (list \"n\" \"n^2\" \"sqrt(n)\")))"
  (let* ((nb-cols (length (car table)))
         (fill-up-spec (lambda (spec-list)
                         (let ((d (- nb-cols (length spec-list))))
                           (if (< 0 d)
                               (append spec-list (make-list d nil))
                             (copy-sequence spec-list)))))
         (col-formatters (funcall fill-up-spec col-formatters))
         (alignments     (funcall fill-up-spec alignments))
         (table (mapcar (lambda (line)
                          (cl-mapcar
                           (lambda (entry formatter)
                             (if formatter (funcall formatter entry) entry))
                           line col-formatters))
                        table))
         (max-col-lengths
          (mapcar
           (lambda (i)
             (apply
              #'max
              (mapcar #'length
                      (mapcar (apply-partially #'nth i)
                              (if header-line
                                  (cons header-line table)
                                table)))))
           (number-sequence 0 (1- nb-cols)))))
    (cl-flet ((format-line (line)
                (let ((i 0))
                  (apply #'concat
                         (cl-mapcar
                          (lambda (entry alignment)
                            (prog1 (concat
                                    (let ((padding (- (nth i max-col-lengths)
                                                      (length entry))))
                                      (pcase alignment
                                        ('right (concat (make-string padding ?\ ) entry))
                                        ((and (or 'center 'centered)
                                              (let left (/ padding 2)))
                                         (concat (make-string left ?\ )
                                                 entry
                                                 (make-string (- padding left) ?\ )))
                                        (_ (concat entry (make-string padding ?\ )))))
                                    (unless (= i (1- nb-cols))
                                      (pcase col-separator
                                        ((and (pred consp) (pred (nth i)) s) s)
                                        ((pred identity) col-separator)
                                        (_ " "))))
                              (cl-incf i)))
                          line alignments)))))
      (concat (and-let* ((header-line)
                         (header-line (format-line header-line)))
                (concat header-line "\n" (make-string (length header-line) ?-) "\n"))
              (mapconcat #'format-line table "\n")))))
